/*
 * Copyright (c) 2009, 2012, 2013, 2014 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin
 *
 * Using FBO as accumulation buffer.
 *
*/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#ifdef __APPLE__
#include <OpenGL/glext.h>
// glfw3.h has GL_GLEXT_LEGACY defined and thus
// doesn't include glext.h automatically
#define glGenVertexArrays glGenVertexArraysAPPLE
#define glBindVertexArray glBindVertexArrayAPPLE
#define GL_RGBA32F GL_RGBA32F_ARB
#else
#include "GL/glew.h"
#endif

#include <GLFW/glfw3.h>

GLdouble width=640.0;
GLdouble height=640.0;   /* window width and height */
GLFWwindow *wd;                   /* GLUT window handle */
GLuint fbod;
int fbwidth, fbheight;
float fbtowd;
GLenum gl_error;

/* Callback functions for GLFW */

/* Called when window is closed, also when 'q' or ESC is hit. */
void
quit(GLFWwindow *wd)
{
  glfwDestroyWindow(wd);
  glfwTerminate();

  exit(0);
}

void
kbd(GLFWwindow* wd, int key, int scancode, int action, int mods)
{
  if (action == GLFW_RELEASE) { // function is called first on GLFW_PRESS.
    return;
  }
  
  switch (key) {
  case GLFW_KEY_ESCAPE:
  case 'Q': // alphabets default to upper case
    quit(wd);
    break;
  default:
    break;
  }

  return;
}

void
err(int errcode, const char* desc)
{
  fprintf(stderr, "%d: %s\n", errcode, desc);

  return;
}

void
disp(void)
{
  GLfloat triad[3][2] = { // (0,0) is the lower left corner
    { 256.0, 384.0 }, 
    { 256.0, 256.0 }, 
    { 384.0, 384.0 }
  }; 
  GLfloat triad2[3][2] = { // (0,0) is the lower left corner
    { 384.0, 384.0 },
    { 256.0, 256.0 }, 
    { 384.0, 256.0 }
  }; 

  GLint maxsize;
  glGetIntegerv(GL_MAX_RENDERBUFFER_SIZE, &maxsize);
  if (fbwidth > maxsize || fbheight > maxsize) {
    cerr << "Framebuffer size > max renderbuffer size" << endl;
    exit(-1);
  }
    
#define NRBUF 2  // number of render buffers
  GLuint rbod[NRBUF];
    
  glBindFramebuffer(GL_FRAMEBUFFER, fbod);

  glGenRenderbuffers(NRBUF, rbod);
  // use first rbo as accumulation buffer
  glBindRenderbuffer(GL_RENDERBUFFER, rbod[0]);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_RGBA32F, fbwidth, fbheight);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "RenderbufferStorage: [" 
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(-1);
  }
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0,
                            GL_RENDERBUFFER, rbod[0]);

  // use second rbo as depth and stencil buffers
  glBindRenderbuffer(GL_RENDERBUFFER, rbod[1]);
  glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH_STENCIL, fbwidth, fbheight);
  glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT,
                            GL_RENDERBUFFER, rbod[1]);
  if ((gl_error = glCheckFramebufferStatus(GL_FRAMEBUFFER)) !=
      GL_FRAMEBUFFER_COMPLETE) {
    cerr << "Framebuffer incomplete [0x" << hex << gl_error << "] "
         << glGetString(gl_error) << endl;
    exit(-1);
  }

  glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);
 
#define NREND 8 // how many samples per pixel
  float weight = 1.0/(float)NREND;

  glEnable(GL_BLEND);
  glBlendColor(0.0, 0.0, 0.0, weight);
  glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE);
  glBlendEquation(GL_FUNC_ADD);

  for (int i = 0; i < NREND; i++) {

    // the first two samples of the first triangle have lighter color 
    if (i == 0) {
      glColor3f(0.12, 0.24, 0.21);
    } else if (i == 1) {
      glColor3f(0.12, 0.24, 0.21);
    } else {
      glColor3f(0.4, 0.8, 0.7);
    }
    glVertexPointer(2, GL_FLOAT, 0, triad); 
    glDrawArrays(GL_TRIANGLES, 0, 3);

    // all samples of the second triangle are of the same color
    glColor3f(0.4, 0.8, 0.7);
    glVertexPointer(2, GL_FLOAT, 0, triad2);
    glDrawArrays(GL_TRIANGLES, 0, 3);

    // so second triangle should be brighter than the first one
  }

  glDisable(GL_BLEND);

  // now blit the "accumulation buffer" to default framebuffer
  glBindFramebuffer(GL_DRAW_FRAMEBUFFER, 0);
  glBindFramebuffer(GL_READ_FRAMEBUFFER, fbod);
  glBlitFramebuffer(0, 0, fbwidth, fbheight, 0, 0, fbwidth, fbheight,
                    GL_COLOR_BUFFER_BIT, GL_NEAREST);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "BlitFramebuffer: [" 
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(-1);
  }
  glBindFramebuffer(GL_FRAMEBUFFER, 0);

  /* GLFW is ALWAYS double buffered, will call glFlush() */
  glfwSwapBuffers(wd);
  
  glDeleteRenderbuffers(NRBUF, rbod);

  return;
}

/* Called when framebuffer is resized, e.g., when window is resized
 * OR when the same size buffer is moved across Retina and non-Retina displays
 * on Mac OS X.
 * NOT called automatically when window is first created.
 * Called by GLFW BEFORE reshape().
*/
void
fbreshape(GLFWwindow *wd, int w, int h)
{
  /* save new framebuffer dimensions */
  fbwidth = w;
  fbheight = h;
  fbtowd = ((float ) fbwidth)/width;

  /* do an orthographic parallel projection with the view volume
     set to first quadrant, fixed to the initial window dimension */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, (float)fbwidth, 0.0, (float)fbheight, -1.f, 1.f);
  
  /* Tell OpenGL to use the whole window for drawing.
     Note that we don't resize the view volume, so
     the viewport will show the whole view volume 
     shrunk/stretched to fit the current view port. */
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  /* return to modelview stack */
  glMatrixMode(GL_MODELVIEW);

  return;
}

/* Called when window is resized.
 * NOT called automatically when window is first created.
 * Called by GLFW AFTER fbreshape().
*/
void
reshape(GLFWwindow *wd, int w, int h)
{
  /* save new screen dimensions */
  width = w;
  height = h;
  fbtowd = ((float ) fbwidth)/width;
  // recompute framebuffer to window ratio, fbreshape is called by GLFW first

  return;
}

void
init_shapes(void)
{
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glEnableClientState(GL_VERTEX_ARRAY); // allow OpenGL server access to client memory
  //glEnableClientState(GL_COLOR_ARRAY);  // allow OpenGL server access to client memory
  glGenFramebuffers(1, &fbod);

  return;
}

int
main(int argc, char *argv[])
{
  glfwSetErrorCallback(err);
  if (!glfwInit()) {
    exit(1);
  }

  /* create the window and its associated OpenGL context */
  wd = glfwCreateWindow(width, height, "Experiment with Framebuffer Object",
                        NULL, NULL);
  if (!wd) {
    glfwTerminate();
    exit(1);
  }
  /* make the window's context the current context */
  glfwMakeContextCurrent(wd);

  /* With Retina display on Mac OS X GLFW's FramebufferSize
     is different from WindowSize */
  glfwGetFramebufferSize(wd, &fbwidth, &fbheight);

  /* shape view port */
  fbreshape(wd, fbwidth, fbheight);

#ifndef __APPLE__
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    /* Problem: glewInit failed, something is seriously wrong. */
    cerr << "Error: " << glewGetErrorString(err) << endl;
  }
  cout <<  "Status: Using GLEW " << glewGetString(GLEW_VERSION) << endl;
#endif /* __APPLE__ */

  /* --- register callbacks with GLFW --- */

  /* register function to handle window resizes */
  /* With Retina display on Mac OS X GLFW's FramebufferSize
     is different from WindowSize */
  glfwSetFramebufferSizeCallback(wd, fbreshape);
  glfwSetWindowSizeCallback(wd, reshape);

  /* register function to handle window close */
  glfwSetWindowCloseCallback(wd, quit);

  /* register function to handle keyboard input */
  glfwSetKeyCallback(wd, kbd);      // general keyboard input

  /* --- draw --- */
  init_shapes();

  do {
    disp();
    glfwWaitEvents();
  } while (!glfwWindowShouldClose(wd));

  exit(0);
}
