/*
 * Copyright (c) 2009, 2012, 2013, 2014 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin
 *
 * VAO without VBO (see vao-vbo.cpp for example of VAO using VBO)
 * and without shaders.  Supported by Mac OS X but not any other
*  graphics card drivers.
 *
*/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
using namespace std;

#ifdef __APPLE__
#include <OpenGL/glext.h>
// glfw3.h has GL_GLEXT_LEGACY defined and thus
// doesn't include glext.h automatically
#define glGenVertexArrays glGenVertexArraysAPPLE
#define glBindVertexArray glBindVertexArrayAPPLE
#else
#include "GL/glew.h"  // must come before the other GL includes
#endif

#include <GLFW/glfw3.h>

static GLenum gl_error;

GLdouble width=640.0;
GLdouble height=480.0;   /* window width and height */
GLFWwindow *wd;                  /* GLUT window handle */
int fbwidth, fbheight;
float fbtowd;

enum { TRIAD=0, QUAD, NumVAOs };

GLuint vaods[NumVAOs];        // variable to hold vao descriptor(s)

GLfloat triad[3][2] = { // (0,0) is the lower left corner
  { 0.0, 64.0 }, 
  { 0.0, 0.0 }, 
  { 64.0, 0.0 } 
}; 
GLfloat quad[4][2] = { // (0,0) is the lower left corner
  { 0.0, 64.0 },  // v0
  { 0.0, 0.0 },   // v1
  { 64.0, 64.0 }, // v2; T0: v0-v1-v2
  { 64.0, 0.0 }   // v3; T1: v2-v1-v3
}; 
GLfloat colors[4][3] = {
  { 1.0, 0.0, 0.0 }, 
  { 0.0, 1.0, 0.0 }, 
  { 0.0, 0.0, 1.0 },
  { 0.0, 1.0, 1.0 }
}; 

/* Callback functions for GLFW */

/* Called when framebuffer is resized, e.g., when window is resized
 * OR when the same size buffer is moved across Retina and non-Retina displays
 * on Mac OS X.
 * NOT called automatically when window is first created.
 * Called by GLFW BEFORE reshape().
*/
void
fbreshape(GLFWwindow *wd, int w, int h)
{
  /* save new framebuffer dimensions */
  fbwidth = w;
  fbheight = h;
  fbtowd = ((float ) fbwidth)/width;

  /* do an orthographic parallel projection with the view volume
     set to first quadrant, fixed to the initial window dimension */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, (float)fbwidth, 0.0, (float)fbheight, -1.f, 1.f);
  
  /* Tell OpenGL to use the whole window for drawing.
     Note that we don't resize the view volume, so
     the viewport will show the whole view volume 
     shrunk/stretched to fit the current view port. */
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  /* return to modelview stack */
  glMatrixMode(GL_MODELVIEW);

  return;
}

/* Called when window is resized.
 * NOT called automatically when window is first created.
 * Called by GLFW AFTER fbreshape().
*/
void
reshape(GLFWwindow *wd, int w, int h)
{
  /* save new screen dimensions */
  width = w;
  height = h;
  fbtowd = ((float ) fbwidth)/width;
  // recompute framebuffer to window ratio, fbreshape is called by GLFW first

  return;
}

/* Called when window is closed,
   also when 'q' or ESC is hit. */
void
quit(GLFWwindow *wd)
{
  glfwDestroyWindow(wd);
  glfwTerminate();

  exit(0);
}

void
kbd(GLFWwindow* wd, int key, int scancode, int action, int mods)
{
  if (action == GLFW_RELEASE) { // function is called first on GLFW_PRESS.
    return;
  }
  
  switch (key) {
  case GLFW_KEY_ESCAPE:
  case 'Q': // alphabets default to upper case
    quit(wd);
    break;
  default:
    break;
  }

  return;
}

void
err(int errcode, const char* desc)
{
  fprintf(stderr, "%d: %s\n", errcode, desc);

  return;
}

void
disp(void)
{
  int x, y;

  glClear(GL_COLOR_BUFFER_BIT);

  for (y=0; y < 10; y++) {
    glLoadIdentity();
    glTranslatef(0.0, y*64.0, 0.0);
    for (x=0; x < 10; x++) {
      glBindVertexArray(vaods[TRIAD]);
      glDrawArrays(GL_TRIANGLES, 0, 3);

      glTranslatef(64.0, 0.0, 0.0);
      glBindVertexArray(vaods[QUAD]);
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);

      glTranslatef(64.0, 0.0, 0.0);
    }
  }

  /* GLFW is ALWAYS double buffered, will call glFlush() */
  glfwSwapBuffers(wd);
  
  return;
}

void
init_shapes(void)
{
  // Set up vertex-array object (vao) descriptors:
  // grab a number of descriptors, for each descriptor, create and bind a vao, 
  // allow OpenGL access to client memory, and set up vertex array pointer.
  /*
     VAO enables switching between multiple sets of vertex arrays
     (analogous to Texture Unit).
     See Vertex-Array Object section in Redbook 7th ed.
     http://www.informit.com/articles/article.aspx?p=1377833&seqNum=8
  */
  glGenVertexArrays(NumVAOs, vaods); // generate an array of n vao descriptors
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "GenVertArr: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glBindVertexArray(vaods[TRIAD]); // create and bind vao(s) to use for subsequent
                                 // gl*VertexAttrib* and gl*Draw* calls
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "BindVertArr triad: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glEnableClientState(GL_VERTEX_ARRAY); // allow OpenGL server access to client memory
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS triad VA: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glVertexPointer(2, GL_FLOAT, 0, (GLvoid *) triad);   // format of data stored in vertex array
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "VertPtr triad: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glEnableClientState(GL_COLOR_ARRAY);  // allow OpenGL server access to client memory
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS triad CA: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glColorPointer(3, GL_FLOAT, 0, (GLvoid *) colors);    // format of data stored in color array
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "ClrPtr triad: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glBindVertexArray(vaods[QUAD]);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "BindVertArr quad: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glEnableClientState(GL_VERTEX_ARRAY);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS quad VA: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glVertexPointer(2, GL_FLOAT, 0, (GLvoid *) quad);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "VertPtr quad: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }


  glEnableClientState(GL_COLOR_ARRAY);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "EnableCS quad CA: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glColorPointer(3, GL_FLOAT, 0, (GLvoid *) colors);
  if ((gl_error = glGetError()) != GL_NO_ERROR) {
    cerr << "ClrPtr quad: [Error "
         << gl_error << "] " << glGetString(gl_error) << endl;
    exit(1);
  }

  glClearColor(1.0, 1.0, 1.0, 1.0);

  return;
}

int
main(int argc, char *argv[])
{

#ifndef __APPLE__
  fprintf(stderr, "%s: only supported on Mac OS X\n", argv[0]);
  exit(1);
#endif

  glfwSetErrorCallback(err);
  if (!glfwInit()) {
    exit(1);
  }

  /* create the window and its associated OpenGL context */
  wd = glfwCreateWindow(width, height, "Experiment with Vertex Array Object WITHOUT Vertex Buffer Object", NULL, NULL);
  if (!wd) {
    glfwTerminate();
    exit(1);
  }
  /* make the window's context the current context */
  glfwMakeContextCurrent(wd);

  /* With Retina display on Mac OS X GLFW's FramebufferSize
     is different from WindowSize */
  glfwGetFramebufferSize(wd, &fbwidth, &fbheight);

  /* shape view port */
  fbreshape(wd, fbwidth, fbheight);

#ifndef __APPLE__
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    /* Problem: glewInit failed, something is seriously wrong. */
    cerr << "Error: " << glewGetErrorString(err) << endl;
  }
  cout <<  "Status: Using GLEW " << glewGetString(GLEW_VERSION) << endl;
#endif /* __APPLE__ */

  /* --- register callbacks with GLFW --- */

  /* register function to handle window resizes */
  /* With Retina display on Mac OS X GLFW's FramebufferSize
     is different from WindowSize */
  glfwSetFramebufferSizeCallback(wd, fbreshape);
  glfwSetWindowSizeCallback(wd, reshape);

  /* register function to handle window close */
  glfwSetWindowCloseCallback(wd, quit);

  /* register function to handle keyboard input */
  glfwSetKeyCallback(wd, kbd);      // general keyboard input

  /* --- draw --- */
  init_shapes();

  do {
    disp();
    glfwWaitEvents();
  } while (!glfwWindowShouldClose(wd));

  exit(0);
}
