/*
 * Copyright (c) 2009, 2012, 2013, 2014 University of Michigan, Ann Arbor.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of Michigan, Ann Arbor. The name of the University 
 * may not be used to endorse or promote products derived from this 
 * software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Sugih Jamin
 *
 * Example of using VBO.
*/
#include <iostream>
#include <stdio.h>
#include <stdlib.h>
using namespace std;

#ifndef __APPLE__
#include "GL/glew.h"
#endif
#include <GLFW/glfw3.h>

enum { TRIAD=0, QUAD, NumVAOs};
long long triadsize, quadsize;

GLdouble width=640.0;
GLdouble height=480.0;   /* window width and height */
GLFWwindow *wd;                   /* GLUT window handle */
int fbwidth, fbheight;
float fbtowd;

GLuint vbods[2];                 // variable to hold vbo descriptor(s)        

/* Callback functions for GLFW */

/* Called when framebuffer is resized, e.g., when window is resized
 * OR when the same size buffer is moved across Retina and non-Retina displays
 * on Mac OS X.
 * NOT called automatically when window is first created.
 * Called by GLFW BEFORE reshape().
*/
void
fbreshape(GLFWwindow *wd, int w, int h)
{
  /* save new framebuffer dimensions */
  fbwidth = w;
  fbheight = h;
  fbtowd = ((float ) fbwidth)/width;

  /* do an orthographic parallel projection with the view volume
     set to first quadrant, fixed to the initial window dimension */
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0.0, (float)fbwidth, 0.0, (float)fbheight, -1.f, 1.f);
  
  /* Tell OpenGL to use the whole window for drawing.
     Note that we don't resize the view volume, so
     the viewport will show the whole view volume 
     shrunk/stretched to fit the current view port. */
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);

  /* return to modelview stack */
  glMatrixMode(GL_MODELVIEW);

  return;
}

/* Called when window is resized.
 * NOT called automatically when window is first created.
 * Called by GLFW AFTER fbreshape().
*/
void
reshape(GLFWwindow *wd, int w, int h)
{
  /* save new screen dimensions */
  width = w;
  height = h;
  fbtowd = ((float ) fbwidth)/width;
  // recompute framebuffer to window ratio, fbreshape is called by GLFW first

  return;
}

/* Called when window is closed,
   also when 'q' or ESC is hit. */
void
quit(GLFWwindow *wd)
{
  glfwDestroyWindow(wd);
  glfwTerminate();

  exit(0);
}

void
kbd(GLFWwindow* wd, int key, int scancode, int action, int mods)
{
  if (action == GLFW_RELEASE) { // function is called first on GLFW_PRESS.
    return;
  }
  
  switch (key) {
  case GLFW_KEY_ESCAPE:
  case 'Q': // alphabets default to upper case
    quit(wd);
    break;
  default:
    break;
  }

  return;
}

void
err(int errcode, const char* desc)
{
  fprintf(stderr, "%d: %s\n", errcode, desc);

  return;
}

void
disp(void)
{
  int x, y;

  glClear(GL_COLOR_BUFFER_BIT);

  for (y=0; y < 10; y++) {
    glLoadIdentity();
    glTranslatef(0.0, y*64.0, 0.0);
    for (x=0; x < 10; x++) {
      glBindBuffer(GL_ARRAY_BUFFER, vbods[TRIAD]); 
      glVertexPointer(2, GL_FLOAT, 0, 0);   // format of data stored in vertex array
      glColorPointer(3, GL_FLOAT, 0, (GLvoid *) triadsize);
                                           // format of data stored in color array
                                           // triadsize specifies the byte offset into
                                           // the vbo's data store where the color array starts
      glDrawArrays(GL_TRIANGLES, 0, 3);
      glTranslatef(64.0, 0.0, 0.0);

      glBindBuffer(GL_ARRAY_BUFFER, vbods[QUAD]); 
      glVertexPointer(2, GL_FLOAT, 0, 0);
      glColorPointer(3, GL_FLOAT, 0, (GLvoid *) quadsize);
      glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
      glTranslatef(64.0, 0.0, 0.0);
    }
  }

  /* GLFW is ALWAYS double buffered, will call glFlush() */
  glfwSwapBuffers(wd);
  
  return;
}

void
init_shapes(void)
{
  GLfloat triad[3][2] = { // (0,0) is the lower left corner
    { 0.0, 64.0 }, 
    { 0.0, 0.0 }, 
    { 64.0, 0.0 } 
  }; 
  GLfloat quad[4][2] = { // (0,0) is the lower left corner
    { 0.0, 64.0 },  // v0
    { 0.0, 0.0 },   // v1
    { 64.0, 64.0 }, // v2; T0: v0-v1-v2
    { 64.0, 0.0 }   // v3; T1: v2-v1-v3
  }; 
  GLfloat colors[4][3] = {
    { 1.0, 0.0, 0.0 }, 
    { 0.0, 1.0, 0.0 }, 
    { 0.0, 0.0, 1.0 },
    { 0.0, 1.0, 1.0 }
  }; 

  triadsize = sizeof(triad);
  quadsize = sizeof(quad);

  /* Set up a server-side vertex buffer object (vbo(s)):
     first generate a descriptor for the vbo(s).  Bind
     the descriptor with a buffer object of specified
     type (GL_ARRAY_BUFFER).  Copy the data 
     over, and specify usage hint OpenGL can use to 
     optimize performance.
  */
  glGenBuffers(2, vbods);      // generate an array of n (=2 here) vbo descriptor

  glBindBuffer(GL_ARRAY_BUFFER, vbods[TRIAD]); // create a new buffer object, of type
                              // GL_ARRAY_BUFFER, attach it to vbos and make
                              // it the current buffer object
  glBufferData(GL_ARRAY_BUFFER, 
               sizeof(triad)+sizeof(colors),
               triad, GL_STATIC_DRAW );
                              // allocate sizeof(triad) OpenGL server
                              // memory, copy data in "triad" to the newly
                              // allocated space. GL_STATIC_DRAW is a performance
                              // hint to OpenGL that the data is for drawing,
                              // and it will not be changed.
  glBufferSubData(GL_ARRAY_BUFFER, sizeof(triad), sizeof(colors), colors);

  glBindBuffer(GL_ARRAY_BUFFER, vbods[QUAD]); 
  glBufferData(GL_ARRAY_BUFFER, sizeof(quad)+sizeof(colors), quad, GL_STATIC_DRAW );
  glBufferSubData(GL_ARRAY_BUFFER, sizeof(quad), sizeof(colors), colors);

  glEnableClientState(GL_VERTEX_ARRAY); // allow OpenGL server access to client memory
  glEnableClientState(GL_COLOR_ARRAY);  // allow OpenGL server access to client memory

  /* clear color buffer to white */
  glClearColor(1.0, 1.0, 1.0, 0.0);
  
  return;
}

int
main(int argc, char *argv[])
{
  glfwSetErrorCallback(err);
  if (!glfwInit()) {
    exit(1);
  }

  /* create the window and its associated OpenGL context */
  wd = glfwCreateWindow(width, height, "Experiment with Vertex Buffer Object",
                        NULL, NULL);
  if (!wd) {
    glfwTerminate();
    exit(1);
  }
  /* make the window's context the current context */
  glfwMakeContextCurrent(wd);

  /* With Retina display on Mac OS X GLFW's FramebufferSize
     is different from WindowSize */
  glfwGetFramebufferSize(wd, &fbwidth, &fbheight);

  /* shape view port */
  fbreshape(wd, fbwidth, fbheight);

#ifndef __APPLE__
  GLenum err = glewInit();
  if (GLEW_OK != err) {
    /* Problem: glewInit failed, something is seriously wrong. */
    cerr << "Error: " << glewGetErrorString(err) << endl;
  }
  cout <<  "Status: Using GLEW " << glewGetString(GLEW_VERSION) << endl;
#endif /* __APPLE__ */

  /* --- register callbacks with GLFW --- */

  /* register function to handle window resizes */
  /* With Retina display on Mac OS X GLFW's FramebufferSize
     is different from WindowSize */
  glfwSetFramebufferSizeCallback(wd, fbreshape);
  glfwSetWindowSizeCallback(wd, reshape);

  /* register function to handle window close */
  glfwSetWindowCloseCallback(wd, quit);

  /* register function to handle keyboard input */
  glfwSetKeyCallback(wd, kbd);      // general keyboard input

  /* --- draw --- */
  init_shapes();

  do {
    disp();
    glfwWaitEvents();
  } while (!glfwWindowShouldClose(wd));

  exit(0);
}
